# -*- coding: utf-8 -*-
# Copyright 2009-2015, Red Hat, Inc.
# License: GPL-2.0+ <http://spdx.org/licenses/GPL-2.0+>
# See the LICENSE file for more details on Licensing

from __future__ import absolute_import
import os.path
import yaml
import pipes

from libtaskotron import logger
from libtaskotron import config
from libtaskotron import remote_exec
from libtaskotron import vm
from libtaskotron.logger import log
import libtaskotron.exceptions as exc


class BaseMinion(object):
    '''Base Minion class that shouldn't be used on its own, it solely initiates inner attributes
    and environment. It also provides method _run() that delegates the task execution to Executor
    over SSH.
    '''

    def __init__(self, task_data, arg_data):
        '''Create a new instance.

        :param dict task_data: data loaded from task formula
        :param dict arg_data: data from command line arguments
        '''

        self.task_data = task_data
        self.arg_data = arg_data
        #: str uuid_dir with path to artifactsdir for storing logs
        self.uuid_dir = self.arg_data['artifactsdir']

        self.user = arg_data.get('user', None)
        self.machine = arg_data.get('machine', None)
        self.port = arg_data.get('port', None)
        self.ssh_privkey = arg_data['ssh_privkey'] or config.get_config().ssh_privkey
        #: int exitcode set by exitcode directive.
        self.exitcode = None
        self.ssh = None

    def _prepare_task(self):
        '''Prepare the environment to run the task.'''

        self.taskdir = config.get_config().client_taskdir

        # FIXME remove this once the vm code is done
        self.ssh.cmd('curl -o /etc/yum.repos.d/tflink-taskotron-fedora-f21.repo '
                     'http://copr-fe.cloud.fedoraproject.org/coprs/tflink/taskotron/'
                     'repo/fedora-21/tflink-taskotron-fedora-21.repo')
        self.ssh.install_pkgs('libtaskotron')

        # configure libtaskotron
        if config.get_config().config_filename:
            self.ssh.put_file(config.get_config().config_filename, '/etc/taskotron/taskotron.yaml')

        # create needed dirs
        self.ssh.mkdir(self.taskdir)

        # patch remote libtaskotron if needed (for dev, shouldn't be used in production)
        if self.arg_data['patch'] is not None:
            self.ssh.put_file(os.path.abspath(self.arg_data['patch']), '%s/%s' % (
                self.taskdir, self.arg_data['patch']))
            self.ssh.install_pkgs('patch')
            self.ssh.cmd('patch -d /usr/lib/python2.7/site-packages/ -p1 -i %s/%s' % (
                self.taskdir, self.arg_data['patch']))

        # put files needed for execution (task, input files, etc)
        self.ssh.put_dir(os.path.dirname(self.arg_data['taskfile']), self.taskdir)

        # need to have default_flow_style false to get valid yaml w/ nested dicts
        self.ssh.write_file(os.path.join(self.taskdir,
                                         os.path.basename(self.arg_data['taskfile'])),
                            yaml.dump(self.task_data, default_flow_style=False))

    def _run(self):
        # execute task
        task_cmd = ['cd', pipes.quote(self.taskdir), '&&',
                    'runtask',
                    '-t', pipes.quote(self.arg_data['type']),
                    '-i', pipes.quote(self.arg_data['item']),
                    '--uuid', pipes.quote(self.arg_data['uuid']),
                    '--local',
                    pipes.quote(os.path.basename(self.arg_data['taskfile']))]

        self.exitcode = self.ssh.cmd(' '.join(task_cmd))

    def _get_output(self):
        '''copy artifacts from vm'''
        try:
            self.ssh.get_dir(self.arg_data['artifactsdir'], self.arg_data['artifactsdir'])
        except exc.TaskotronRemoteError, e:
            log.info(e)

    def execute(self):
        '''This method has to be implemented by classes that inherits from :class:`.BaseMinion`.
        It should contain environment initialization, namely set `self.ssh` to properly
        instantiated :class:`.ParamikoWrapper` and cleanup (if necessarily). Moreover, the order of
        private methods has to be: `_prepare_task()`, `_run()` and `_get_output()`.
        '''
        raise exc.TaskotronNotImplementedError()


class PersistentMinion(BaseMinion):
    '''Minion class that connects to alredy running machine using SSH and runs the task there.
    '''

    def execute(self):
        '''Init environment and connect to remote machine for task execution.'''

        filelog_path = os.path.join(self.uuid_dir, 'taskotron-initiator.log')
        logger.add_filehandler(filelog_path=filelog_path, remove_mem_handler=True)

        stdio_filename = os.path.join(self.uuid_dir, 'taskotron-stdio.log')

        log.info("Running task over SSH on machine '%s'...", self.machine)

        with remote_exec.ParamikoWrapper(self.machine, self.port, self.user,
                                         self.ssh_privkey, stdio_filename) as self.ssh:
            self._prepare_task()
            self._run()
            self._get_output()


class DisposableMinion(BaseMinion):
    '''Minion class that creates a disposable client, connects to it over SSH and runs the task
    there.
    '''

    def execute(self):
        '''Init environment and connect to disposable client for task execution.'''

        filelog_path = os.path.join(self.uuid_dir, 'taskotron-initiator.log')
        logger.add_filehandler(filelog_path=filelog_path, remove_mem_handler=True)

        stdio_filename = os.path.join(self.uuid_dir, 'taskotron-stdio.log')

        execution_ok = False
        try:
            log.info("Running task on a disposable client using libvirt...")

            log.info("Creating VM for task")
            task_vm = vm.TestCloudMachine(self.arg_data['uuid'])
            task_vm.prepare()
            self.machine = task_vm.ipaddr
            self.port = 22
            self.user = task_vm.username

            log.info("Running Task over SSH on machine %s" % task_vm.instancename)

            task_vm.wait_for_port(self.port)

            with remote_exec.ParamikoWrapper(self.machine, self.port, self.user,
                                             self.ssh_privkey, stdio_filename) as self.ssh:
                    self._prepare_task()
                    self._run()
                    self._get_output()
            execution_ok = True
        finally:
            if self.arg_data['no_destroy']:
                log.info("--no-destroy mode was set from command line, not tearing VM "
                         "down for task %s" % self.arg_data['uuid'])
            else:
                try:
                    task_vm.teardown()
                except exc.TaskotronRemoteError, e:
                    if execution_ok:
                        raise e
                    else:
                        # a different exception is already being raised, don't override it
                        log.exception(e)
