# -*- coding: utf-8 -*-
# Copyright 2009-2015, Red Hat, Inc.
# License: GPL-2.0+ <http://spdx.org/licenses/GPL-2.0+>
# See the LICENSE file for more details on Licensing

'''Tools for remote execution primary for disposable clients'''

from __future__ import absolute_import
import os
import os.path
import sys
import socket
from stat import S_ISDIR
from time import sleep

import paramiko

from .logger import log
from . import exceptions as exc

from libtaskotron.directives import exitcode_directive
from libtaskotron.file_utils import Tee


class ParamikoWrapper(object):
    '''Wrapper for SSH communication using paramiko library'''

    # timeout for network operations in seconds (900 seconds = 15 minutes)
    TIMEOUT = 900

    def __init__(self, hostname, port, username, key_filename, stdio_filename=None):
        self.ssh = None
        self.sftp = None
        self.hostname = hostname
        self.port = port
        self.username = username
        self.key_filename = key_filename
        self.stdio_filename = stdio_filename
        self.outstream = Tee(sys.stdout)

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def __str__(self):
        return '<%s: %s@%s:%s>' % (self.__class__.__name__, self.username,
                                   self.hostname, self.port)

    def connect(self):
        '''Connect to a machine over ssh. Open sftp channel and, if applicable, file that
           stdout/err from the machine will be saved to.

           :raise TaskotronRemoteError: when the connection does not succeed'''

        self.ssh = paramiko.SSHClient()
        # accept unknown hosts
        self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        log.debug('Connecting to remote host: %s@%s:%s', self.username, self.hostname, self.port)

        try:
            self.ssh.connect(self.hostname,
                             port=self.port,
                             username=self.username,
                             # we use empty string by default (no special key), but paramiko
                             # requires receiving None in that case
                             key_filename=self.key_filename or None)

            self.sftp = self.ssh.open_sftp()
            self.sftp.get_channel().settimeout(self.TIMEOUT)

            if self.stdio_filename is not None:
                try:
                    f = open(self.stdio_filename, 'w')
                    self.outstream.add(f)
                except IOError, e:
                    log.warning('Could not open %s. Falling back to writing vm\'s output '
                                'to stdout only.', self.stdio_filename)
        except paramiko.BadHostKeyException, e:
            raise exc.TaskotronRemoteError('Server\'s (%s@%s:%s) hostkey could not be verified: %s'
                                           % (self.username, self.hostname, self.port, str(e)))
        except paramiko.AuthenticationException, e:
            raise exc.TaskotronRemoteError('Authentication to %s@%s:%s failed: %s' %
                                           (self.username, self.hostname, self.port, str(e)))
        except (paramiko.SSHException, socket.error), e:
            raise exc.TaskotronRemoteError('Could not connect to %s@%s:%s: %s' %
                                           (self.username, self.hostname, self.port, str(e)))
        except IOError as e:
            # let's hope this does not occur in more situations
            raise exc.TaskotronRemoteError("The private key '%s' could not be read: %s" %
                                           (self.key_filename, str(e)))

    def close(self):
        '''Close open connections and files.'''

        self.outstream.close()
        self.sftp.close()
        self.ssh.close()

    def cmd(self, cmd):
        '''Execute a command.

        :param str cmd: A command to be executed
        :returns: returncode of the command
        :raise TaskotronRemoteError: If the command has non-zero return code and it isn't a code of
                                     exitcode directive failure. Or if the remote hasn't sent any
                                     data for ``TIMEOUT`` seconds.
        '''

        log.debug('Running command on remote host: %s', cmd)

        stdin, stdout, stderr = self.ssh.exec_command(cmd, timeout=self.TIMEOUT)

        # stdout.channel represents channel for both stdout and stderr
        channel = stdout.channel

        # there seems to be no way to tell whether the remote is alive,
        # so just use the counter
        # https://phab.qadevel.cloud.fedoraproject.org/T593
        alive_counter = 0
        while not channel.exit_status_ready():
            if channel.recv_ready():
                data = channel.recv(1024)
                self.outstream.write(data)
                alive_counter = 0
            if channel.recv_stderr_ready():
                data = channel.recv_stderr(1024)
                self.outstream.write(data)
                alive_counter = 0

            sleep(0.1)
            alive_counter += 0.1

            if alive_counter >= self.TIMEOUT:
                raise exc.TaskotronRemoteError('Has not received any output from the remote '
                                               '%s@%s:%s for %d seconds',
                                               (self.username, self.hostname, self.port,
                                                self.TIMEOUT))

        retcode = channel.recv_exit_status()

        # read the rest of data that was left in the buffers
        self.outstream.write(stdout.read())
        self.outstream.write(stderr.read())

        if retcode != 0 and retcode != exitcode_directive.FAILURE:
            raise exc.TaskotronRemoteError('Command "%s" on %s@%s exited with code %s' %
                                           (cmd, self.username, self.hostname, retcode))
        return retcode

    def install_pkgs(self, pkgs):
        '''Install packages via dnf.

        :param list pkgs: A list or a string separated by commas of packages to be installed

        :raise TaskotronRemoteError: If the command has non-zero return code
        '''

        if isinstance(pkgs, basestring):
            pkgs = pkgs.split(',')

        log.debug('Installing packages on remote host: %s', ','.join(pkgs))
        self.cmd('dnf -y install %s' % ' '.join(pkgs))

    def write_file(self, remote_path, data, overwrite=True):
        '''Write data to a remote file.

        :param str remote_path: A path to the remote file
        :param str data: Data to be written
        :param bool overwrite: Whether to overwrite remote path. Default is True.

        :raise TaskotronRemoteError: If data could not be written
        '''

        if self._remote_file_exists(remote_path) and not overwrite:
            log.info('Remote path %s already exists, not overwriting.', remote_path)
            return

        log.debug('Writing data to %s', remote_path)
        try:
            with self.sftp.open(remote_path, 'w') as remote_file:
                remote_file.write(data)
        except (socket.error, IOError), e:
            raise exc.TaskotronRemoteError('Could not write data to %s@%s:%s: %s' %
                                           (self.username, self.hostname, remote_path, e.strerror))

    def put_file(self, local_path, remote_path, overwrite=True):
        '''Copy a file to a remote path.

        :param str local_path: A path to the local file
        :param str remote_path: A path to the remote file
        :param bool overwrite: Whether to overwrite remote path. Default is True.

        :return: an SFTPAttributes object containing attributes about the given file,
                 if successful. None otherwise.

        :raise TaskotronRemoteError: If the file could not be copied
        '''

        if self._remote_file_exists(remote_path) and not overwrite:
            log.info('Remote path %s already exists, not overwriting.', remote_path)
            return

        log.debug('Putting file (%s) on to remote host (%s)', local_path, remote_path)
        try:
            return self.sftp.put(local_path, remote_path)
        except (socket.error, IOError), e:
            raise exc.TaskotronRemoteError('Could not put file %s to %s@%s:%s: %s' %
                                           (local_path, self.username, self.hostname,
                                            remote_path, e.strerror))

    def get_file(self, remote_path, local_path):
        '''Get a file from a remote path.

        :param str remote_path: A path to the remote file
        :param str local_path: A path to the local file

        :raise TaskotronRemoteError: If the file could not be downloaded
        '''

        log.debug('Getting file from remote host %s to %s', remote_path, local_path)
        try:
            self.sftp.get(remote_path, local_path)
        except (socket.error, IOError), e:
            raise exc.TaskotronRemoteError('Could not copy file %s@%s:%s to %s: %s' %
                                           (self.username, self.hostname, remote_path,
                                            local_path, e.strerror))

    def _remote_file_exists(self, remote_path):
        try:
            path_stats = self.sftp.stat(remote_path)
            if path_stats is not None:
                return True
        except IOError:
            pass

        return False

    def _remote_isdir(self, remote_path):
        try:
            return S_ISDIR(self.sftp.stat(remote_path).st_mode)
        except IOError:
            return False

    def mkdir(self, name, mode=777):
        '''Create a directory on a remote machine.

        :param str name: A name of the directory
        :param int mode: Permissions (posix-style, decimal) of the directory,
                         default is 777

        :raise TaskotronRemoteError: If the directory could not be created
        '''

        self.cmd('mkdir -p -m %d %s' % (mode, name))

    def put_dir(self, local_path, remote_path, overwrite=True):
        '''Copy a directory to a remote path.

        :param str remote_path: A path to the remote directory
        :param str local_path: A path to the local directory
        :param bool overwrite: Whether to overwrite remote path. Default is True.

        :raise TaskotronRemoteError: If the directory could not be uploaded
        '''

        if self._remote_file_exists(remote_path) and not overwrite:
            log.info('Remote path %s already exists, not overwriting.', remote_path)
            return

        try:
            files = os.listdir(local_path)
        except OSError, e:
            raise exc.TaskotronRemoteError('Could not put dir %s to %s@%s:%s: %s' %
                                           (local_path, self.username, self.hostname,
                                            remote_path, e.strerror))

        self.mkdir(remote_path)

        for file_ in files:
            remote_file = os.path.join(remote_path, file_)
            local_file = os.path.join(local_path, file_)
            if os.path.isdir(local_file):
                if not self._remote_isdir(remote_file):
                    self.sftp.mkdir(remote_file)
                self.put_dir(local_file, remote_file)
            else:
                self.put_file(local_file, remote_file)

    def get_dir(self, remote_path, local_path):
        '''Get a directory from a remote path.

        :param str remote_path: A path to the remote directory
        :param str local_path: A path to the local directory

        :raise TaskotronRemoteError: If the directory could not be downloaded
        '''

        try:
            files = self.sftp.listdir(remote_path)
        except IOError, e:
            raise exc.TaskotronRemoteError('Could not copy dir %s@%s:%s to %s: %s' %
                                           (self.username, self.hostname, remote_path,
                                            local_path, e.strerror))

        for file_ in files:
            remote_file = os.path.join(remote_path, file_)
            local_file = os.path.join(local_path, file_)
            if self._remote_isdir(remote_file):
                os.mkdir(local_file)
                self.get_dir(remote_file, local_file)
            else:
                self.get_file(remote_file, local_file)
